//
// General functions shared by both Avi2Cvc (Video) and Bmp2Cvi (Image) converters
//
module utils;

import std.math;		// abs

@safe:

//
// Ordered Dithering
// =================
// Moved the pattern array out of the function to reduce allocations and make the GC happier.
// This should make things a bit quicker too!
//

const ubyte[][] pattern=[
[ 0, 32,  8, 40,  2, 34, 10, 42],   //* 8x8 Bayer ordered dithering  */
[48, 16, 56, 24, 50, 18, 58, 26],   //* pattern.  Each input pixel   */
[12, 44,  4, 36, 14, 46,  6, 38],   //* is scaled to the 0..63 range */
[60, 28, 52, 20, 62, 30, 54, 22],   //* before looking in this table */
[ 3, 35, 11, 43,  1, 33,  9, 41],   //* to determine the action.     */
[51, 19, 59, 27, 49, 17, 57, 25],
[15, 47,  7, 39, 13, 45,  5, 37],
[63, 31, 55, 23, 61, 29, 53, 21] ];

void ordered(ubyte[] gray,const int w, const int h)
{ 
	ubyte col;
	int pos,prow,pcol;
	
	foreach(y;0..h)
	{
		foreach(x;0..w)
		{
			col=gray[pos]>>2;
			if (col<pattern[pcol][prow]) 
			{
				gray[pos]=3;
				gray[pos]-=cast(ubyte)(col>>4); // 3- col>>4
			}
			else
			{
				gray[pos]=0;
			}

			pcol++;
			if (pcol>7){pcol=0;}
			
			pos++;	
		}

		prow++;
		if (prow>7){prow=0;}
	}
}

//
// Error Diffusion
// ===============
// Also moved the Cybiko Colour array (cycol) out of the function to reduce allocations and make the GC happier.
// Again, this should make things a little quicker too!
//

const ubyte[] cycol=[255,171,85,0]; //0,85,171,255
	
void diffuse(ubyte[] gray,const int w, const int h)
{ 
	int pos,tc;
	ubyte closest,no;
	int carry;
	int best,dif;
	
	foreach(y;0..h)
	{
		carry=0;
		foreach(x;0..w)
		{
			best=999;
			closest=0;
			tc=gray[pos]+carry;
			for(no=0;no<4;no++)
			{
				dif=abs(cycol[no]-tc);
				if (dif<best)
				{
					best=dif;
					closest=no;
					carry=tc-cycol[no];
				}
			}
			gray[pos]=closest;
			pos++;
		}	
	}
}

//
// Converts 24bit colour to 8bit gray (and flips)
//

void grey(ubyte[] bmp,ubyte[]gray,const int w, const int h)
{
	ubyte r,g,b;
	int c,pos;
	
	foreach(y;0..h)
	{
		c=((h-1)-y)*(w*3);	// flip image
	
		foreach(x;0..w)
		{
			r=bmp[c];
			g=bmp[c+1];
			b=bmp[c+2];
			gray[pos]=(r+g+b)/3;
			//gray[pos]=(r >>2)+ (g>>1) + (b>>2);
		
			c+=3;
			pos++;
		}	
	}
}



@safe void twobit(ubyte[] gray,const int w, const int h){
	foreach(ref g;gray){
		g=3-(g>>6);
	}
}


//
// Converts frame from 8bit (gray) to 2bit
//

void frameconv(ubyte[] gray,ubyte[] output,const int w, const int h){
	int c,pos;
	
	foreach(y;0..h){
		foreach(x;0..(w/4)){
			ubyte pa=gray[c];
			ubyte pb=gray[c+1];
			ubyte pc=gray[c+2];
			ubyte pd=gray[c+3];
			c=c+4;
			
			output[pos]=cast(ubyte)(pa<<6);
			output[pos]+=cast(ubyte)(pb<<4);
			output[pos]+=cast(ubyte)(pc<<2);
			output[pos]+=cast(ubyte)(pd);
			pos++;
		}	
	}
}
